/* ------------------------------------------------------------------------- */
/*
 *  hcss.h
 *
 *  Copyright (c) 2007 "clown". All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - No names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Last-modified: Tue 24 Apr 2007 12:55:16 JST
 */
/* ------------------------------------------------------------------------- */
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <iterator>
#include <locale>
#include <stdexcept>

#include "clx/remove.h"
#include "clx/strip.h"

#ifndef CLX_HCSS_H
#define CLX_HCSS_H

// syntax error messages
#define EHCSS_BLOCK		"expected '}' before EOF"
#define EHCSS_COMMBEGIN	"expected '/*' before '*/' token"
#define EHCSS_COMMEND	"expected '*/' before EOF"
#define EHCSS_SELECTOR	"expected selector before '{' token"
#define EHCSS_DECLS		"expected ';' token"
#define EHCSS_DECL		"expected property and ':' before ';' token"
#define EHCSS_PROPERTY	"expected property before ':' token"
#define EHCSS_VALUE		"expected value before ';' token"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  css_syntax_error
	/* --------------------------------------------------------------------- */
	class css_syntax_error : public std::runtime_error {
	public:
		typedef size_t size_type;
		typedef char char_type;
		typedef std::string string_type;
		
		explicit css_syntax_error(size_type n, const string_type& what_arg) :
			std::runtime_error(what_arg), line_(n) {}
		
		size_type line() { return line_; }
		
	private:
		size_type line_;
	};
	
	/* --------------------------------------------------------------------- */
	//  cssdecl
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	struct cssdecl {
		typedef CharT char_type;
		typedef std::basic_string<CharT, Traits> string_type;
		
		string_type property;
		string_type value;
		
		cssdecl() : property(), value() {}
		
		explicit cssdecl(const string_type& prop, const string_type& val) :
			property(prop), value(val) {}
		
		explicit cssdecl(const char_type* prop, const char_type* val) :
			property(prop), value(val) {}
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_hcss
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_hcss {
	public:
		typedef CharT char_type;
		typedef size_t size_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef std::vector<string_type> selector_list;
		typedef cssdecl<CharT, Traits> declaration_type;
		typedef std::vector<declaration_type> declaration_list;
		
		basic_hcss() : sels_(), v_(), sp_(), line_(1) {}
		
		template <class InputIterator>
		explicit basic_hcss(InputIterator first, InputIterator last, const std::locale& loc = std::locale()) :
			sels_(), v_(), sp_(), line_(1) {
			this->assign(first, last, loc);
		}
		
		explicit basic_hcss(const char_type* filename, const std::locale& loc = std::locale()) :
			sels_(), v_(), sp_(), line_(1) {
			this->assign(filename, loc);
		}
		
		explicit basic_hcss(const string_type& filename, const std::locale& loc = std::locale()) :
			sels_(), v_(), sp_(), line_(1) {
			this->assign(filename, loc);
		}
		
		// operator
		template <class InputIterator>
		basic_hcss& assign(InputIterator first, InputIterator last, const std::locale& loc = std::locale()) {
			string_type tmp;
			while (first != last) {
				if (this->is_comment_begin(first, tmp)) {
					this->skip_comment(first, last, tmp);
					continue;
				}
				
				if (this->is_comment_end(first, tmp)) throw css_syntax_error(line_, EHCSS_COMMBEGIN);
				else if (*first == '{') this->push_entry(tmp, loc);
				else if (*first == '}') this->pop_selector(tmp, loc);
				else if (*first == ':') this->push_declaration(tmp, first, last, loc);
				else if (*first == ';') throw css_syntax_error(line_, EHCSS_DECL);
				else if (*first == '\n') line_++;
				else tmp += *first;
				
				first++;
			}
			
			if (!sels_.empty()) throw css_syntax_error(line_, EHCSS_BLOCK);
			return *this;
		}
		
		basic_hcss& assign(const char_type* filename, const std::locale& loc = std::locale()) {
			std::basic_ifstream<CharT, Traits> fs(filename);
			if (fs.fail()) {
				std::basic_stringstream<CharT, Traits> msg;
				msg << filename << ": no such file or directory";
				throw std::runtime_error(msg.str());
			}
			
			std::istreambuf_iterator<char_type> first(fs);
			std::istreambuf_iterator<char_type> last;
			this->assign(first, last, loc);
			
			fs.close();
			return *this;
		}
		
		basic_hcss& assign(const string_type& filename, const std::locale& loc = std::locale()) {
			return this->assign(filename.c_str(), loc);
		}
		
		// accessor
		size_type size() const { return v_.size(); }
		const string_type& selector(size_type i) const { return v_.at(i).first; }
		const declaration_list& declarations(size_type i) const { return v_.at(i).second; }
		
	private:
		typedef std::pair<string_type, declaration_list> decl_block;
		typedef std::vector<decl_block> container;
		
		selector_list sels_;
		container v_;
		
		std::vector<size_type> sp_;
		size_type line_;
		
		template <class InputIterator>
		bool is_comment_begin(const InputIterator& pos, string_type& s) {
			if (!s.empty() && s.at(s.size() - 1) == '/' && *pos == '*') return true;
			return false;
		}
		
		template <class InputIterator>
		bool is_comment_end(const InputIterator& pos, string_type& s) {
			if (!s.empty() && s.at(s.size() - 1) == '*' && *pos == '/') return true;
			return false;
		}
		
		template <class InputIterator>
		void skip_comment(InputIterator& first, InputIterator last, string_type& s) {
			s.erase(s.size() - 1);
			
			first++;
			do {
				if (*first == '\n') line_++;
				else if (*first == '*' && (++first == last || *first == '/')) break;
			} while (++first != last);
			
			if (first == last) throw css_syntax_error(line_, EHCSS_COMMEND);
			first++;
		}
		
		void push_entry(string_type& s, const std::locale& loc) {
			if (remove_if(s, is_space(loc)).empty()) {
				throw css_syntax_error(line_, EHCSS_SELECTOR);
			}
			sels_.push_back(s);
			
			std::basic_stringstream<CharT, Traits> ss;
			for (size_type i = 0; i < sels_.size(); i++) {
				if (sels_.at(i) != "*" && sels_.at(i) != "body") {
					ss << sels_.at(i);
					if (i + 1 < sels_.size()) ss << ' ';
				}
			}
			
			decl_block tmp;
			if (ss.str().empty()) tmp.first = sels_.at(sels_.size() - 1);
			else tmp.first = ss.str();
			v_.push_back(tmp);
			sp_.push_back(v_.size() - 1);
			
			s.erase();
		}
		
		void pop_selector(string_type& s, const std::locale& loc) {
			sels_.pop_back();
			sp_.pop_back();
			s.erase();
		}
		
		template <class InputIterator>
		void push_declaration(string_type& prop, InputIterator& first, InputIterator last, const std::locale& loc) {
			if (strip(prop, loc).empty()) throw css_syntax_error(line_, EHCSS_PROPERTY);
			
			first++;
			string_type tmp;
			while (first != last) {
				if (this->is_comment_begin(first, tmp)) {
					this->skip_comment(first, last, tmp);
					continue;
				}
				
				if (this->is_comment_end(first, tmp)) throw css_syntax_error(line_, EHCSS_COMMBEGIN);
				else if (*first == ':') throw css_syntax_error(line_, EHCSS_DECLS);
				else if (*first == ';' || *first == '}') break;
				else if (*first == '\n') line_++;
				tmp += *first++;
			}
			if (first == last) throw css_syntax_error(line_, EHCSS_BLOCK);
			
			if (strip(tmp, loc).empty()) throw css_syntax_error(line_, EHCSS_VALUE);
			declaration_type decl(prop, tmp);
			v_.at(sp_.back()).second.push_back(decl);
			
			if (*first == '}') this->pop_selector(prop, loc);
			else prop.erase();
		}
	};
	
	typedef basic_hcss<char> hcss;
}

#endif // CLX_HCSS_H
